import click
from hich.commands.pairs import pairs
from hich.commands.fasta import fasta
from hich.commands.report import report
from hich.commands.matrix import matrix

@click.group()
def hich():
    """CLI tools to process Hi-C data
    
    Version: 0.2.4
    """
    pass

hich.add_command(fasta)
hich.add_command(pairs)
hich.add_command(report)
hich.add_command(matrix)

if __name__ == "__main__":
    hich()


from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
from Bio.Restriction import RestrictionBatch
from pathlib import Path
import smart_open_with_pbgzip
from smart_open import smart_open
from typing import *
import itertools
import polars as pl

protocols = {
    'HiC 3.0': ['DpnII', 'DdeI'],
    'Hi-C 3.0': ['DpnII', 'DdeI'],
    'Arima Genome-Wide HiC+': ['DpnII', 'HinfI'],
    'Arima': ['DpnII', 'HinfI'],
    'Phase Proximo 2021+ Plant': ['DpnII'],
    'Phase Plant': ['DpnII'],
    'Phase Proximo 2021+ Animal': ['DpnII'],
    'Phase Animal': ['DpnII'],
    'Phase Proximo 2021+ Microbiome': ['Sau3AI', 'MluCI'],
    'Phase Microbiome': ['Sau3AI', 'MluCI'],
    'Phase Proximo 2021+ Human': ['DpnII'],
    'Phase Human': ['DpnII'],
    'Phase Proximo 2021+ Fungal': ['DpnII'],
    'Phase Fungal': ['DpnII']
}

def unique_enzymes(enzymes_and_protocols):
    """Extract unique entries from flattened iterable"""
    protocols_upper = {k.upper(): v for k, v in protocols.items()}
    enzymes = [protocols_upper.get(it.upper(), [it]) for it in enzymes_and_protocols]
    return list(set(itertools.chain.from_iterable(enzymes)))

def sequences(fasta: Path | str) -> Generator[SeqRecord, None, None]:
    """Yields sequences from fasta file (may be compressed)
    """
    handle = smart_open(fasta, "rt")
    yield from SeqIO.parse(handle, "fasta")

def pos_to_bed(chrom: str, pos: List[int], end: int, start: int = 1) -> pl.DataFrame:
    """Convert list of positions to a BED dataframe
    """
    # Ensure starts contains zero and not the last value
    # Ensure ends contains the last value and not zero
    starts = sorted(list(set([p for p in [start, *pos] if p != end])))
    ends = sorted(list(set([p for p in [*pos, start + end] if p != start]))) 
    chroms = [chrom]*len(starts)

    return (
        pl.DataFrame({
            "chrom": chroms,
            "start": starts,
            "end": ends
        })
    )


def digest_re(fasta: Path | str, enzymes_and_protocols: List[str]):
    enzymes = unique_enzymes(enzymes_and_protocols)
    try:
        restriction_batch = RestrictionBatch(enzymes)
    except Exception as e:
        print(f"Failed to create RestrictionBatch using enzymes list {enzymes}")
        print(e)

    for sequence in sequences(fasta):
        digest = restriction_batch.search(sequence.seq)
        pos = list(itertools.chain.from_iterable(digest.values()))
        yield pos_to_bed(sequence.id, pos, len(sequence.seq))




import click
import smart_open_with_pbgzip
from smart_open import smart_open
import duckdb
import polars as pl
import shutil
from pathlib import Path
from hich.pairs.convert import walk_files, rename, reorganize
from hich.pairs.pairssql import PairsSQL
import os
from parse import parse
import numpy as np

@click.group()
def pairs():
    pass

@pairs.command()
@click.option("--in-format", type = click.Choice(["autodetect", "duckdb", "pairs", "parquet"]), default = "autodetect", help = "Input file format")
@click.option("--out-format", type = click.Choice(["autodetect", "duckdb", "pairs", "parquet"]), default = "autodetect", help = "Output file format")
@click.option("--in-pattern", type = str, default = None, help = "Python parse format for extracting names from original partitioned file.")
@click.option("--out-pattern", type = str, default = None, help = "python parse format for creating new filename. Output files should NOT be created in OUT_PATH directory.")
@click.option("--sql", type = str, default = None, help = "SQL to run on input file before partition.")
@click.option("--squote", default = "\"", help = "Replace this string with a single quote ' in the sql string")
@click.option("--unlink", is_flag = True, default = False, help = "Delete original partitioned file if renaming.")
@click.option("--memory-limit", type = int, default = None, help = "DuckDB memory limit in GB")
@click.option("--threads", type = int, default = None, help = "DuckDB thread limit")
@click.argument("in-path")
@click.argument("out-path")
@click.argument("partition_by", nargs=-1)
def partition(in_format, out_format, in_pattern, out_pattern, sql, squote, unlink, memory_limit, threads, in_path, out_path, partition_by):
    """Split a .pairs-like file into multiple .pairs-like outputs

    \b
    IN_PATH: Path to input file to be partitioned (.pairs, .pairssql, .pairsparquet)
    OUT_PATH: Location where partitioned output files will be initially generated
    [PARTITION_BY]: Columns to partition the output; one output file generated per combination of values in these columns

    By default, all files are stored as a pairsparquet file named "data_0.parquet" in a partition-specific subdirectory of OUT_PATH. Subdirectories reflect a tree structure based on values of PARTITION_BY. The names of the first tier of subdirectories are values of the first column in PARTITION_BY, the second tier reflects values in the second column in PARTITION_BY, etc.

    \b
    Examples:
    Split to per-chromosome pairsparquet files in the directory structure output_dir/chrom1={chrom1_val}/chrom2={chrom2_val}/data_0.parquet:
        "hich pairs partition all_cells.pairs output_dir chrom1 chrom2"
    Convert outputs to .pairs format files named ./results/{chrom1_val}_{chrom2_val}.pairs:
        "hich pairs partition --in-pattern "output_dir/chrom1={chrom1}/chrom2={chrom2}/data_0.parquet" --out-pattern "results/{chrom1}_{chrom2}.pairs all_cells.pairs" output_dir chrom1 chrom2"
    Split by same vs. different chromosomes when that was not already labeled in the .pairs file:
        "hich pairs partition --sql "ALTER TABLE pairs ADD COLUMN same_chrom BOOLEAN; UPDATE pairs SET same_chrom = (chrom1 = chrom2)" all_cells.pairs output_dir same_chrom
    """
    if not partition_by:
        raise ValueError(f"No column names to partition {in_path} by were submitted.")

    try:
        db = PairsSQL().open(in_path, in_format)
        db.set_memory_limit(memory_limit)
        db.set_threads(threads)
        
    except Exception as e:
        print(f"Failed to open {in_path} with format {in_format}")
        raise e

    try:
        if sql:
            if squote:
                sql = sql.replace(squote, "'")
            db.conn.execute(sql)
    except Exception as e:
        print(f"Preliminary SQL query failed: {sql}")
        raise e

    try:
        db.partition_by(out_path, partition_by)
    except Exception as e:
        print(f"Failed to partition {in_path} by {partition_by} in output directory {out_path} ")
        raise e
    
    try:
        reorganize(out_path, in_pattern, out_pattern, in_format, out_format, unlink)
    except Exception as e:
        raise e

@pairs.command()
@click.option("--in-format", type = click.Choice(["autodetect", "duckdb", "pairs"]), default = "autodetect", help = "Input file format")
@click.option("--out-format", type = click.Choice(["autodetect", "duckdb", "parquet", "pairs", "tsv", "csv", "sql"]), default = "autodetect", help = "Output file format")
@click.option("--squote", default = "\"", help = "Replace this string with a single quote ' in the sql string")
@click.option("--out-path", default = None, help = "If supplied, changes are rewritten to this file, otherwise to stdout")
@click.option("--print-sql", default = False, is_flag = True, help = "Print SQL instead of running it")
@click.option("--memory-limit", type = int, default = None, help = "DuckDB memory limit in GB")
@click.option("--threads", type = int, default = None, help = "DuckDB thread limit")
@click.argument("sql")
@click.argument("in-path")
def sql(in_format, out_format, squote, out_path, print_sql, memory_limit, threads, sql, in_path):
    """Run a DuckDB SQL query on a .pairs-like file

    The 4DN .pairs format is ingested to '.pairssql' format using DuckDB, which has a `pairs` table having the same columns and names as the original .pairs file. Column types are autodetected through a full scan of the entire .pairs file. If outputting to .pairs, the header will be updated with any changed column names. If outputting to Parquet or DuckDB, the output will store the original .pairs header, either as a parquet kv metadata value "header" or the DuckDB table "metadata". The header will lack the #columns: line as this is generated on the fly when outputting to .pairs from the pairs table columns. 

    \b
    SQL: The DuckDB SQL query to run over file after ingesting to DuckDB. May also be a path to a file containing an SQL command.
    IN_PATH: Path to input file; use /dev/stdin to read from stdin

    \b
    Examples:
    \b
    Extract the substring of the readID column prior to the first ':' character and set as the value of the cellID column
        hich pairs sql "ALTER TABLE pairs ADD COLUMN cellID VARCHAR; UPDATE pairs SET cellID = regexp_extract(readID, \"(.*):(.*)\", 1);" no_cellID.pairs cellID_labeled.pairs
    Add a log10 distance strata with null values for transchromosomal or zero-distance contacts
        hich pairs sql "ALTER TABLE pairs ADD COLUMN distance INTEGER; UPDATE pairs SET distance = ROUND(LOG10(pos2 - pos1)) WHERE chrom1 == chrom2 AND pos1 != pos2;"
    Drop contacts mapping to different chromosomes
        hich pairs sql "DELETE FROM pairs WHERE chrom1 != chrom2;"
    Count number of contacts mapping to different distance strata:
        hich pairs sql "CREATE TEMP TABLE pairs_counts AS SELECT CAST(ROUND(LOG10(pos2-pos1)) AS INTEGER) A
S distance, COUNT(*) AS count FROM pairs WHERE pos1 != pos2 AND chrom1 == chrom2 GROUP BY distance; DROP TABLE pairs; CREATE TABLE pairs AS SELECT * FROM pairs_counts;"
    """
    try:
        # Load SQL from file
        sql_path = Path(sql)
        if sql_path.exists():
            sql = smart_open(sql_path).read()
    except:
        pass
    if squote:
        sql = sql.replace(squote, "'")
    if print_sql:
        print(sql)
    else:
        db = PairsSQL.open(in_path, in_format)
        db.set_memory_limit(memory_limit)
        db.set_threads(threads)
        try:
            if sql:
                db.conn.execute(sql)
        except Exception as e:
            print(f"SQL command failed on {in_path}:\n{sql}")
            print(e)
        db.write(out_path, out_format)

@pairs.command()
@click.option("--in-format", type = click.Choice(["autodetect", "duckdb", "pairs"]), default = "autodetect", help = "Input file format")
@click.option("--out-format", type = click.Choice(["autodetect", "duckdb", "parquet", "pairs", "sql"]), default = "autodetect", help = "Output file format")
@click.option("--idx1", default = "idx1", show_default=True, help = "Label of first index")
@click.option("--start1", default = "start1", show_default=True, help = "Label of second BED interval")
@click.option("--end1", default = "end1", show_default=True, help = "Label of first BED interval")
@click.option("--idx2", default = "idx2", show_default=True, help = "Label of second BED index")
@click.option("--start2", default = "start2", show_default=True, help = "Label of second BED start")
@click.option("--end2", default = "end2", show_default=True, help = "Label of second BED end")
@click.option("--batch-size", default = 1000, show_default=True, help = "Number of blocks of 1024 rows at a time to treat as batch size")
@click.option("--memory-limit", type = int, default = None, help = "DuckDB memory limit in GB")
@click.option("--threads", type = int, default = None, help = "DuckDB thread limit")
@click.argument("in_path_pairs")
@click.argument("in_path_bed")
@click.argument("out_path")
def bin(in_format, out_format, idx1, start1, end1, idx2, start2, end2, batch_size, memory_limit, threads, in_path_pairs, in_path_bed, out_path):
    """Bin contacts using a nonuniform partition

    IN_PATH_PAIRS: A .pairs-like file to compute intersections on the ends of its contacts
    IN_PATH_BED: A BED file defining the partition
    OUT_PATH: .pairs-like filename where results are saved (updates IN_PATH_PAIRS if they are the same)
    """   
    same_file = Path(in_path_pairs) == Path(out_path)
    assert not same_file or in_format == out_format, "Cannot change format while keeping same filename."

    in_db = PairsSQL.open(in_path_pairs, in_format)
    in_db.set_memory_limit(memory_limit)
    in_db.set_threads(threads)

    out_db = PairsSQL(path=":memory:")
    out_db.set_memory_limit(memory_limit)
    out_db.set_threads(threads)
    out_db.add_metadata(in_db.header)


    # Validate that bed file has no gaps or overlaps
    bed = duckdb.read_csv(in_path_bed, names=["chrom", "start", "end"]).pl().sort("chrom", "start")
    row1_iter = bed.iter_rows()
    row2_iter = bed.iter_rows()
    next(row2_iter)
    for row1, row2 in zip(row1_iter, row2_iter):
        chroms_match = row1[0] == row2[0]
        gap_or_overlap = row2[1] != row1[2]
        if chroms_match and gap_or_overlap:
            raise ValueError(f"Bed file {in_path_bed} does not define a partition. Gap at {row1} and {row2}.")
        

    def intersection_labels(pairs_chunk, bed, chrom, pos_col, idx_label, start_label, end_label):
        bed_chrom = bed.filter(pl.col("chrom") == chrom)

        if bed_chrom is None or len(bed_chrom) == 0:
            idx = [0]*len(pairs_chunk)
            start = [-1]*len(pairs_chunk)
            end = [-1]*len(pairs_chunk)
        elif not bed_chrom["start"][0] == 1:
            raise ValueError(f"For {chrom}, BED file is missing start position 1")
        elif pairs_chunk[pos_col].max() >= 1 + bed_chrom["end"][-1]:
            raise ValueError(f"For {chrom}, found a position {pairs_chunk[pos_col].max()} higher than max BED file partition ending at {bed_chrom['end'][-1]}")
        else:
            idx = np.searchsorted(bed_chrom["end"], pairs_chunk[pos_col], side = "right")
            start = bed_chrom["start"][idx]
            end = bed_chrom["end"][idx]

        return {idx_label: idx, start_label: start, end_label: end}

    for chunk in in_db.iter_chroms(batch_size):
        chrom1 = chunk["chrom1"][0]
        frag1 = intersection_labels(chunk, bed, chrom1, "pos1", idx1, start1, end1)
        
        chrom2 = chunk["chrom2"][0]
        frag2 = intersection_labels(chunk, bed, chrom2, "pos2", idx2, start2, end2)
        tagged_chunk = pl.concat([pl.from_pandas(chunk), pl.DataFrame(frag1), pl.DataFrame(frag2)], how = "horizontal")


        tagged_chunk_schema = tagged_chunk.filter(False)
        if Path(in_path_pairs) == Path(out_path):
            table = "pairs_temp"
        else:
            table = "pairs"
        out_db.conn.execute(
f"""
CREATE TABLE IF NOT EXISTS {table} AS
SELECT *
FROM tagged_chunk_schema;

INSERT INTO {table}
SELECT * FROM tagged_chunk;
"""
        )

    if same_file:
        out_db.conn.execute(
"""
DROP TABLE IF EXISTS pairs;

CREATE TABLE pairs AS SELECT * FROM pairs_temp;

DROP TABLE IF EXISTS pairs_temp;
"""
        )

    out_db.write(out_path, out_format)



"""Parse 4DN .pairs format headers
https://github.com/4dn-dcic/pairix/blob/master/pairs_format_specification.md

Works with compressed files using smart_open. The pairtools suite saves using pbgzip, which smart_open can't use by default. The smart_open_with_pbgzip makes both smart_open and DuckDB use pbgzip.
"""

import smart_open_with_pbgzip
from smart_open import smart_open
from pathlib import Path
from dataclasses import dataclass
from typing import *

@dataclass
class PairsHeader:
    """Parser for 4DN .pairs format header

    https://github.com/4dn-dcic/pairix/blob/master/pairs_format_specification.md

    path: path to the .pairs file
    lines: list of raw header lines
    columns: list of column names'
    misformatted: true if most recently parsed header was misformatted
    data: extracted from non-columns header lines
    
    Methods:
        __init__ -- calls parse by default
        parse -- extract header from file or raw text

    Properties and setters
        parse_header_to_columns -- convert #columns: line in header to self.columns
        parse_columns_to_header -- convert self.columns to #columns: line in header
        text -- raw text of header
        non_columns_text -- raw text of header, any #columns: line removed
        non_columns_lines -- raw text of header in list of separate lines (has setter, does not reparse)
        last_line_columns -- returns True if last header line starts with #columns:, false otherwise 
    """
    path: Path | None = None
    lines: List[str] | None = None
    columns: List[str] = None
    misformatted: bool = None
    data: Dict[str, Any] = dict

    def __init__(self, 
                 data: Path | str, 
                 from_text: bool = False, 
                 columns: List[str] | None = None, 
                 parse: bool = True, 
                 parse_path: bool | None = None, 
                 replace_columns: bool | None = True,
                 replace_data: bool = True
        ):
        """
        Arguments:
            data: Either raw header text (parsed this way if from_text is true) or a path to a 4DN .pairs file (from_text is false)
            from_text: Determines whether to parse data as raw text or path
            columns: List of predetermined columns. Either appended to lines (final line is not a columns line) or replaces final columns line
            parse: Immediately parse the file or text 
            parse_path: If parse is called, enforces whether it parses self.path or self.lines. If None, uses what's available, raises exception if both/None are
            replace_columns: If columns is set, then if replace_columns is True, columns is replaced if a #columns: line is in the header
        """
        if from_text:
            # rstrip the line because a trailing \n on the last line will be set by split as a trailing empty line 
            self.lines = [line + "\n" for line in data.rstrip().split("\n")]
            
            if columns:
                columns_line = " ".join(["#columns:", *columns]) + "\n"
                if self.lines[-1].startswith("#columns:"):
                    self.lines[-1] = columns_line
                else:
                    self.lines.append(columns_line)
        else:
            self.path = Path(data)
        
        if (self.lines or self.path.exists()) and parse:
            self.parse(parse_path, replace_columns, replace_data)

    def parse(self, parse_path: bool = None, replace_columns: bool = False, replace_data: bool = False):
        """Parse 4DN .pairs header or raw text lines

        If parsing .pairs file, file may be compressed or non-compressed. Works with compressed files using smart_open. The pairtools suite saves using pbgzip, which smart_open can't use by default. The smart_open_with_pbgzip makes both smart_open and DuckDB use pbgzip.
        
        If misformatted, PairsHeader.misformatted is set to True.
        Extracts columns to PairsHeader.columns.
        Stores other lines as strings, lists or a dict (for "chromsize") in data.

        Arguments:
            store_lines: Store the raw lines of the header in PairsHeader.header
            parse_path: If true, parses path. If false, parse lines. If None, parse whichever exists (raise if both exist).
            replace_columns: If columns is set, then if replace_columns is True, columns is replaced if a #columns: line is in the header
            replace_data: If data set set, then if replace_data is True, old data is cleared before update.

        File format description:
        https://github.com/4dn-dcic/pairix/blob/master/pairs_format_specification.md
        """
        parse_path = self.update_parse_path_HELPER(parse_path)
        update_columns = (self.columns and replace_columns) or not self.columns
        
        error_preamble = self.get_error_preamble_HELPER(parse_path)
        header = self.get_header_HELPER(parse_path, error_preamble)
        self.columns_HELPER(header, update_columns)
        self.data_HELPER(header, replace_data)

    ###################################
    #   Getters and setters
    ###################################

    @property
    def text(self):
        """Get text of all lines"""
        return "".join(self.lines) if self.lines else None
    
    @property
    def non_columns_text(self):
        """Get text of all lines except columns, if present"""
        return "".join(self.non_columns_lines) if self.lines else None
    
    @property
    def last_line_columns(self) -> bool:
        """Returns whether the last header line starts with #columns:"""
        return self.lines[-1].startswith("#columns:") if self.lines else False

    @property
    def columns_text(self):
        """Get text of columns line"""
        if self.columns:
            joined_columns = " ".join(["#columns:", *self.columns])
        if self.lines[-1].startswith("#columns:"):
            columns_text = self.lines[-1]
            assert not self.columns or columns_text == joined_columns, f"Both self.columns and columns line were set, but did not match.\nColumns line: {columns_line}.\nself.columns: {self.columns}"
        elif self.columns:
            columns_text = joined_columns
        else:
            columns_text = None
        return columns_text

    @property
    def non_columns_lines(self):
        """Get lines except final #columns: line"""
        non_columns_lines = None
        if self.lines:
            non_columns_lines = self.lines[:-1] if self.last_line_columns else self.lines
        return non_columns_lines

    @non_columns_lines.setter
    def non_columns_lines(self, non_columns_lines: List[str]):
        """Set lines other than #columns:"""
        if self.last_line_columns:
            self.lines = non_columns_lines + self.lines[-1]
        else:
            self.lines = non_columns_lines

    def parse_header_to_columns(self):
        """Set self.columns based on #columns: line in header"""
        assert self.last_line_columns, f"Tried to parse header to columns but last line does not start with '#columns:'\nLast line: {self.lines[-1]}"
        self.columns = self.lines[-1].split()[1:]

    def parse_columns_to_header(self):
        """Append/replace #columns: line in header based on self.columns"""
        assert self.columns, f"Tried to parse columns to header but self.columns is empty or None"
        new_columns_line = " ".join(["#columns:", *self.columns])
        if self.last_line_columns:
            self.columns[-1] = new_columns_line
        else:
            self.columns.append(new_columns_line)

    #####################################
    # Helper methods
    #####################################

    def update_parse_path_HELPER(self, parse_path: bool):
        """Set parse_path, if None, to true/false based on whether self.path or self.lines is available"""
        if parse_path is None:
            if self.path and not self.lines:
                parse_path = True
            elif self.lines and not self.path:
                parse_path = False
            elif self.path and self.lines:
                raise ValueError("Called parse on PairsHeader with parse_path set to None, which is ambiguous. Set to True or False when both path and lines are set before call to parse.")
        if not (self.path or self.lines):
            raise ValueError("Called parse on PairsHeader where neither path nor lines are available, so there is nothing to parse.")
        return parse_path
    
    def load_pairs_header_HELPER(self, path: Path | str) -> List[str]:
        """Load list of lines from 4DN .pairs file."""
        path = Path(path)
        error_preamble = f"Attempted to parse pairs header from {path} with absolute path {path.absolute()}"
        assert path.exists(), f"{error_preamble}, but file does not exist."
        self.lines = []
        try:
            with smart_open(path, "rt", encoding="utf-8") as f:
                for line in f:
                    if line.startswith("#"):
                        self.lines.append(line)
                    else:
                        break
        except Exception as e:
            print(f"{error_preamble}, but caught an exception while opening and parsing the file: {e}")
            raise

    def get_header_HELPER(self, parse_path: bool, error_preamble: str):
        """Load header from file or self.lines"""
        if parse_path:
            self.load_pairs_header_HELPER(self.path)
        header = self.lines
        if not header:
            self.misformatted = True
            raise ValueError(f"{error_preamble}, but data did not begin with a header where lines start with #.")
        return header

    def get_error_preamble_HELPER(self, parse_path: bool):
        """Get start of error message depending on if parsing from path or text lines"""
        if parse_path:
            error_preamble = f"Attempted to parse pairs header from {self.path} with absolute path {self.path.absolute()}"
        else:
            error_preamble = f"Attempted to parse pairs header from lines"
        return error_preamble

    def columns_HELPER(self, header: List[str], update_columns: bool):
        """Build self.columns from header"""
        if self.last_line_columns and update_columns:
            self.parse_header_to_columns()
        if not self.columns:
            self.misformatted = True

    def data_HELPER(self, header: List[str], replace_data: bool):
        """Build self.data dict from header"""
        if replace_data:
            self.data = {}

        for line in header[:-1]:
            trimmed = line.lstrip("#").strip()
            if not trimmed:
                continue
            split_line = line.split()

            # Trim leading '#' and trailing ":"
            field = split_line[0]
            value = split_line[1:]

            if field == "chromsize":
                self.data.setdefault("chromsize", {})
                chrom, size = value
                self.data["chromsize"][chrom] = size
            else:
                entry = " ".join(value)
                if field in self.data:
                    if isinstance(self.data[field], str):
                        self.data[field] = [self.data[field]]
                    self.data[field].append(entry)
                else:
                    self.data[field] = entry